package mark0;

import simudyne.core.abm.Action;
import simudyne.core.abm.Agent;
import simudyne.core.annotations.Variable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.stream.Stream;

public class Economy extends Agent<Globals> {
    @Variable
    public double savings;

    public double deficit;

    public double totalFirmCash;

    @Variable
    public double unemployment;

    public double employment;

    @Variable
    public double averagePrice;

    public double prevAveragePrice;

    @Variable
    public double inflation;

    @Variable
    public double averageWage;

    @Variable
    public double totalConsumptionBudget;

    public HashMap<Long, Double> wageMap;

    public HashMap<Long, Double> priceMap;

    public HashMap<Long, Double> productivityMap;

    public HashMap<Long, HealthyFirmAccount> healthyFirmAccountMap;

    public HashMap<Long, Double> indebtedFirmsMap;

    public HashMap<Long, BailoutPackage> bailedOutFirmsMap;

    public HashMap<Long, Double> cashMap;

    public ArrayList<Long> bankruptFirmsArray;

    public Economy() {
        savings = 0;
        deficit = 0;
        totalFirmCash = 0;
        unemployment = 0;
        employment = 1;
        averagePrice = 0;
        inflation = 0;
        averageWage = 1;
        totalConsumptionBudget = 0;
        wageMap = new HashMap<>();
        priceMap = new HashMap<>();
        productivityMap = new HashMap<>();
        healthyFirmAccountMap = new HashMap<>();
        indebtedFirmsMap = new HashMap<>();
        bailedOutFirmsMap = new HashMap<>();
        cashMap = new HashMap<>();
        bankruptFirmsArray = new ArrayList<>();
    }

    public static Action<Economy> initSavings() {
        return Action.create(Economy.class, economy -> {
            double totalCash = economy.getMessagesOfType(Messages.CashMessage.class).stream().mapToDouble(msg ->
                    msg.cash).sum();
            economy.savings = economy.getGlobals().N_f - totalCash;
        });
    }

    public static Action<Economy> initWageMap() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.WageMessage.class).forEach(wageMessage -> {
                economy.wageMap.put(wageMessage.getSender(), wageMessage.wage);
            });
        });
    }

    public static Action<Economy> initPriceMap() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.PriceMessage.class).forEach(priceMessage -> {
                economy.priceMap.put(priceMessage.getSender(), priceMessage.price);
            });
        });
    }

    public static Action<Economy> initProductivityMap() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.ProductivityMessage.class).forEach(productivityMessage -> {
                economy.productivityMap.put(productivityMessage.getSender(), productivityMessage.productivity);
            });
        });
    }

    public static Action<Economy> receivePriceMessage() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.PriceMessage.class).forEach(priceMessage -> {
                economy.priceMap.replace(priceMessage.getSender(), priceMessage.price);
            });
        });
    }

    public static Action<Economy> receiveWageMessages() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.WageMessage.class).forEach(wageMessage -> {
                economy.wageMap.replace(wageMessage.getSender(), wageMessage.wage);
            });
        });
    }

    public static Action<Economy> receiveProductivity() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.ProductivityMessage.class).forEach(productivityMessage -> {
                economy.productivityMap.replace(productivityMessage.getSender(), productivityMessage.productivity);
            });
        });
    }

    public static Action<Economy> calcEmploymentUnemployment() {
        return Action.create(Economy.class, economy -> {
            double totalOutput = economy.getMessagesOfType(Messages.OutputMessage.class).stream().mapToDouble(msg ->
                    msg.output).sum();
            economy.unemployment = 1 - (totalOutput / (economy.getGlobals().mu * economy.getGlobals().N_f));
            economy.employment = 1 - economy.unemployment;
        });
    }

    public static Action<Economy> sendEmploymentUnemployment() {
        return Action.create(Economy.class, economy -> {
           economy.getLinks(Links.EconomyToFirmLink.class).send(Messages.EmploymentUnemploymentMessage.class,
                   (msg, link) -> {
                        msg.employment = economy.employment;
                        msg.unemployment = economy.unemployment;
                   });
        });
    }

    public static Action<Economy> registerUnemployment() {
        return Action.create(Economy.class, economy -> economy.getDoubleAccumulator("Unemployment").add(economy.unemployment));
    }

    public static Action<Economy> calcAveragePrice() {
        return Action.create(Economy.class, economy -> {
            economy.prevAveragePrice = economy.averagePrice;
            Stream<Messages.PriceMessage> messageStreamOutput = economy.getMessagesOfType(Messages.PriceMessage.class).stream();
            double totalOutput = messageStreamOutput.mapToDouble(msg -> msg.output).sum();
            Stream<Messages.PriceMessage> messageStreamPrices = economy.getMessagesOfType(Messages.PriceMessage.class).stream();
            double sumPricesFactored = messageStreamPrices.mapToDouble(msg -> msg.output * msg.price).sum();
            economy.averagePrice = sumPricesFactored / totalOutput;
        });
    }

    public static Action<Economy> sendAveragePrice() {
        return Action.create(Economy.class, economy -> {
            economy.getLinks(Links.EconomyToFirmLink.class).send(Messages.AveragePriceMessage.class,
                    (msg, link) -> {
                        msg.averagePrice = economy.averagePrice;
                    });
        });
    }

    public static Action<Economy> calcAverageWage() {
        return Action.create(Economy.class, economy -> {
            Stream<Messages.WageMessage> messageStreamOutput = economy.getMessagesOfType(Messages.WageMessage.class).stream();
            double totalOutput = messageStreamOutput.mapToDouble(msg -> msg.output).sum();
            Stream<Messages.WageMessage> messageStreamWages = economy.getMessagesOfType(Messages.WageMessage.class).stream();
            double sumWagesFactored = messageStreamWages.mapToDouble(msg -> msg.output * msg.wage).sum();
            economy.averageWage = sumWagesFactored / totalOutput;
        });
    }

    public static Action<Economy> calcInflation() {
        return Action.create(Economy.class, economy -> {
            economy.inflation = (economy.averagePrice - economy.prevAveragePrice) / economy.prevAveragePrice;
            economy.inflation = economy.inflation / economy.getGlobals().gamma_p;
        });
    }

    public static Action<Economy> calcAvailableWorkers() {
        return Action.create(Economy.class, economy -> {
            double dNom = economy.productivityMap.keySet().stream().mapToDouble(firmID ->
                    economy.productivityMap.get(firmID) * Math.exp((economy.getGlobals().beta *
                            economy.wageMap.get(firmID))/economy.averageWage)).sum();

            economy.wageMap.keySet().forEach(firmID -> {
                double availableWorkersNom = Math.exp((economy.getGlobals().beta * economy.wageMap.get(firmID))/ economy.averageWage);
                double availableWorkers = (availableWorkersNom / dNom) * economy.getGlobals().N_f * economy.unemployment;
                economy.getLinksTo(firmID).send(Messages.AvailableWorkersMessage.class, (msg, link) ->
                        msg.availableWorkers = availableWorkers);
            });
        });
    }

    public static Action<Economy> calcTotalConsumptionBudget() {
        return Action.create(Economy.class, economy -> {
            Stream<Messages.WageMessage> messageStream = economy.getMessagesOfType(Messages.WageMessage.class).stream();
            double sumWagesFactored = messageStream.mapToDouble(msg -> msg.output * msg.wage).sum();
            economy.totalConsumptionBudget = economy.getGlobals().C * (Math.max(economy.savings, 0) + sumWagesFactored);
        });
    }

    public static Action<Economy> calcDemand() {
        return Action.create(Economy.class, economy -> {
            double dNom = economy.productivityMap.keySet().stream().mapToDouble(firmID ->
                    economy.productivityMap.get(firmID) * Math.exp(-(economy.getGlobals().beta *
                            economy.priceMap.get(firmID))/economy.averagePrice)).sum();

            economy.priceMap.keySet().forEach(firmID -> {
                double priceNom = economy.productivityMap.get(firmID) * Math.exp(-(economy.getGlobals().beta * economy.priceMap.get(firmID))/ economy.averagePrice);
                double demand =(economy.totalConsumptionBudget / economy.priceMap.get(firmID)) * (priceNom / dNom);
                economy.getLinksTo(firmID).send(Messages.DemandMessage.class, (msg, link) ->
                        msg.demand = demand);
            });
        });
    }

    public static Action<Economy> receiveProfit() {
        return Action.create(Economy.class, economy -> {
            double totalProfit = economy.getMessagesOfType(Messages.ProfitMessage.class).stream().mapToDouble(msg -> msg.profit).sum();
            economy.savings += totalProfit;
        });
    }

    public static Action<Economy> receiveHealthyFirmAccounts() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.HealthyFirmAccountMessage.class).forEach(msg -> {
                economy.healthyFirmAccountMap.put(msg.getSender(), msg.healthyFirmAccount);
            });
        });
    }

    public static Action<Economy> receiveIndebtedFirmDebt() {
        return Action.create(Economy.class, economy -> {
            economy.getMessagesOfType(Messages.BailoutRequestMessage.class).forEach(msg -> {
                economy.indebtedFirmsMap.put(msg.getSender(), msg.debt);
            });
        });
    }

    public static Action<Economy> checkDefaults() {
        return Action.create(Economy.class, economy -> {
           economy.indebtedFirmsMap.forEach((firmID, debt) -> {
               if (economy.healthyFirmAccountMap.size() > 0) {
                   ArrayList<Long> healthyFirmIDs = new ArrayList<>(economy.healthyFirmAccountMap.keySet());
                   int idx = economy.getPrng().generator.nextInt(economy.healthyFirmAccountMap.size());
                   long healthyFirmID = healthyFirmIDs.get(idx);
                   if ((economy.getPrng().uniform(0, 1).sample() < 1 - economy.getGlobals().f) & (economy.healthyFirmAccountMap.get(healthyFirmID).deposits > -debt)) {
                       economy.healthyFirmAccountMap.get(firmID).updateDeposits(debt);
                       economy.bailedOutFirmsMap.put(firmID, new BailoutPackage(economy.healthyFirmAccountMap.get(firmID).price, economy.healthyFirmAccountMap.get(firmID).wage));
                   }
                   else {
                       economy.deficit -= debt;
                       economy.bankruptFirmsArray.add(firmID);
                   }
               }
               else {
                   economy.deficit -= debt;
                   economy.bankruptFirmsArray.add(firmID);
               }
           });
           economy.indebtedFirmsMap.clear();
        });
    }

    public static Action<Economy> sendHealthyFirmDeposits() {
        return Action.create(Economy.class, economy -> {
           economy.healthyFirmAccountMap.forEach((firmID, healthyFirmAccount) -> {
               economy.getLinksTo(firmID, Links.EconomyToFirmLink.class).send(Messages.DepositsMessage.class, (msg, link) -> {
                   msg.cash = healthyFirmAccount.deposits;
               });
           });
           economy.healthyFirmAccountMap.clear();
        });
    }

    public static Action<Economy> sendBailoutPackages() {
        return Action.create(Economy.class, economy -> {
            economy.bailedOutFirmsMap.forEach((firmID, bailoutPackage) -> {
                economy.getLinksTo(firmID, Links.EconomyToFirmLink.class).send(Messages.BailoutPackageMessage.class, (msg, link) -> {
                    msg.price = bailoutPackage.price;
                    msg.wage = bailoutPackage.wage;
                });
            });
            economy.bailedOutFirmsMap.clear();
        });
    }

    public static Action<Economy> sendBankruptcyMessages() {
        return Action.create(Economy.class, economy -> {
            economy.bankruptFirmsArray.forEach(firmID -> {
                economy.getLinksTo(firmID, Links.EconomyToFirmLink.class).send(Messages.BankruptcyMessage.class);
                });
            economy.bankruptFirmsArray.clear();
            });
    }

    public static Action<Economy> receiveDebt() {
        return Action.create(Economy.class, economy -> {
            double totalDebt = economy.getMessagesOfType(Messages.DebtMessage.class).stream().mapToDouble(msg -> msg.debt).sum();
            economy.deficit += totalDebt;
        });
    }

    public static Action<Economy> receiveCash() {
        return Action.create(Economy.class, economy -> {
            Stream<Messages.CashMessage> totalCashStream = economy.getMessagesOfType(Messages.CashMessage.class).stream();
            double totalCash = totalCashStream.mapToDouble(msg -> msg.cash).sum();
            economy.totalFirmCash += totalCash;
            Stream<Messages.CashMessage> cashMapStream = economy.getMessagesOfType(Messages.CashMessage.class).stream();
            cashMapStream.forEach(msg -> {
                economy.cashMap.put(msg.getSender(), msg.cash);
            });
        });
    }

    public static Action<Economy> distributeDebt() {
        return Action.create(Economy.class, economy -> {
            if (economy.deficit > economy.savings){
                economy.deficit -= economy.savings;
                economy.savings = 0;
                economy.cashMap.forEach((firmID, cash) -> {
                    economy.getLinksTo(firmID, Links.EconomyToFirmLink.class).send(Messages.CashMessage.class, (msg, link) -> {
                        msg.cash = cash - ((cash / economy.totalFirmCash) * economy.deficit);
                    });
                });
            }
            else {
                economy.savings -= economy.deficit;
            }
            economy.cashMap.clear();
            economy.totalFirmCash = 0;
            economy.deficit = 0;
        });
    }

}
