package mark0;

import simudyne.core.abm.Action;
import simudyne.core.abm.Agent;
import simudyne.core.annotations.Variable;

public class Firm extends Agent<Globals> {
    @Variable
    public double wage;

    @Variable
    public double price;

    @Variable
    public double output;

    @Variable
    public double deposits;

    @Variable
    public boolean productivity;

    public double employment;

    public double unemployment;

    public double averagePrice;

    public double availableWorkers;

    public double demand;

    public double profit;

    public boolean healthy;

    public double dividend;

    public double debt;

    public double cash;

    public Firm() {
        wage = 1.0;
        price = 1.0;
        output = 1.0;
        deposits = wage * output * 2.0;
        productivity = true;
        availableWorkers = 0.0;
        demand = 1.0;
        profit = (price * Math.min(output, demand)) - (wage * output);
        healthy = true;
        dividend = 0.0;
        debt = 0.0;
    }

    public static Action<Firm> initVariables() {
        return Action.create(Firm.class, firm -> {
            firm.price = 1.0 + (0.2 * (firm.getPrng().uniform(0, 1).sample() - 0.5));
            firm.output = (firm.getGlobals().mu * (1.0 + (0.2 * (firm.getPrng().uniform(0, 1).sample() - 0.5)))) / 2.0;
            firm.deposits = firm.wage * firm.output * 2.0 * firm.getPrng().uniform(0, 1).sample();
            firm.profit = (firm.price * Math.min(firm.output, firm.demand)) - (firm.wage * firm.output);
        });
    }

    public static Action<Firm> sendDeposits() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.CashMessage.class,
                    (cashMessage, firmToEconomyLink) -> {
                        cashMessage.cash = firm.deposits;
                    });
        });
    }

    public static Action<Firm> sendOutput() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.OutputMessage.class,
                    (outputMessage, firmToEconomyLink) -> {
                        outputMessage.output = firm.output;
                    });
        });
    }

    public static Action<Firm> sendPrice() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.PriceMessage.class,
                    (priceMessage, firmToEconomyLink) -> {
                        priceMessage.output = firm.output;
                        priceMessage.price = firm.price;
                        if (firm.productivity)
                            priceMessage.productivity = 1.0d;
                        else
                            priceMessage.productivity = 0.0d;
                    });
        });
    }

    public static Action<Firm> sendWage() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.WageMessage.class,
                    (wageMessage, firmToEconomyLink) -> {
                        wageMessage.output = firm.output;
                        wageMessage.wage = firm.wage;
                        if (firm.productivity)
                            wageMessage.productivity = 1.0d;
                        else
                            wageMessage.productivity = 0.0d;
                    });
        });
    }

    public static Action<Firm> sendProductivity() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.ProductivityMessage.class,
                    (productivityMessage, firmToEconomyLink) -> {
                        if (firm.productivity)
                            productivityMessage.productivity = 1;
                        else
                            productivityMessage.productivity = 0;
                    });
        });
    }

    public static Action<Firm> receiveAvailableWorkers() {
        return Action.create(Firm.class, firm -> {
            if (firm.getMessagesOfType(Messages.AvailableWorkersMessage.class).stream().count() > 0) {
                firm.availableWorkers = firm.getMessageOfType(Messages.AvailableWorkersMessage.class).availableWorkers;
            }
        });
    }

    public static Action<Firm> receiveEmploymentUnemployment() {
        return Action.create(Firm.class, firm -> {
            if (firm.getMessagesOfType(Messages.EmploymentUnemploymentMessage.class).stream().count() > 0) {
                firm.employment = firm.getMessageOfType(Messages.EmploymentUnemploymentMessage.class).employment;
                firm.unemployment = firm.getMessageOfType(Messages.EmploymentUnemploymentMessage.class).unemployment;
            }
        });
    }

    public static Action<Firm> receiveAveragePrice() {
        return Action.create(Firm.class, firm -> {
            if (firm.getMessagesOfType(Messages.AveragePriceMessage.class).stream().count() > 0) {
                firm.averagePrice = firm.getMessageOfType(Messages.AveragePriceMessage.class).averagePrice;
            }
        });
    }

    public static Action<Firm> calculateStrategy() {
        return Action.create(Firm.class, firm -> {
            if (firm.productivity) {
                if (firm.output < firm.demand) {
                    if (firm.profit > 0) {
                        firm.wage = firm.wage * (1.0d + firm.getGlobals().gamma_w * firm.employment * firm.getPrng().uniform(0, 1).sample());
                        firm.wage = Math.min(firm.wage, (firm.price * Math.min(firm.demand, firm.output))/firm.output);
                    }
                    firm.output = firm.output + Math.min(firm.getGlobals().etta_plus * (firm.demand - firm.output), firm.getGlobals().mu * firm.availableWorkers);
                    if (firm.price < firm.averagePrice) {
                        firm.price = firm.price * (1.0d + firm.getGlobals().gamma_p * firm.getPrng().uniform(0, 1).sample());
                    }
                } else if (firm.output > firm.demand) {
                    if (firm.profit < 0) {
                        firm.wage = firm.wage * (1.0d - firm.getGlobals().gamma_w * firm.unemployment * firm.getPrng().uniform(0, 1).sample());
                    }
                    firm.output = Math.max(0.0d, firm.output - (firm.getGlobals().etta_minus * (firm.output - firm.demand)));

                    if (firm.price > firm.averagePrice) {
                        firm.price = firm.price * (1.0d - firm.getGlobals().gamma_p * firm.getPrng().uniform(0, 1).sample());
                    }
                }
            }
        });
    }

    public static Action<Firm> receiveDemand() {
        return Action.create(Firm.class, firm -> {
            if (firm.getMessagesOfType(Messages.DemandMessage.class).stream().count() > 0) {
                firm.demand = firm.getMessageOfType(Messages.DemandMessage.class).demand;
            }
        });
    }

    public static Action<Firm> doAccounting() {
        return Action.create(Firm.class, firm -> {
            firm.healthy = false;
            if (firm.productivity) {
                firm.profit = (firm.price * Math.min(firm.demand, firm.output)) - (firm.output * firm.wage);
                firm.deposits += firm.profit;
                if (firm.profit > 0 & firm.deposits > 0) {
                    firm.dividend = firm.getGlobals().delta * firm.profit;
                    firm.deposits -= firm.dividend;
                }
                if (firm.deposits > firm.getGlobals().Theta * firm.output * firm.wage) {
                    firm.healthy = true;
                }
            }
        });
    }

    public static Action<Firm> sendProfit() {
        return Action.create(Firm.class, firm -> {
            if (firm.productivity) {
                firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.ProfitMessage.class, (msg, link) -> {
                    msg.profit = -firm.profit + firm.dividend;
                });
            }
            firm.dividend = 0;
        });
    }

    public static Action<Firm> sendBailoutRequest() {
        return Action.create(Firm.class, firm -> {
            if (firm.productivity & (firm.deposits < -firm.getGlobals().Theta * firm.output * firm.wage))
                firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.BailoutRequestMessage.class, (msg, link) -> {
                    msg.debt = firm.deposits;
                });
        });
    }

    public static Action<Firm> sendHealthyFirmAccount() {
        return Action.create(Firm.class, firm -> {
            if (firm.healthy)
                firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.HealthyFirmAccountMessage.class, (msg, link) -> {
                    msg.healthyFirmAccount = new HealthyFirmAccount(firm.deposits, firm.price, firm.wage);
                });
        });
    }

    public static Action<Firm> receiveHealthyFirmDeposits() {
        return Action.create(Firm.class, firm -> {
            if (firm.getMessagesOfType(Messages.DepositsMessage.class).stream().count() > 0) {
                firm.deposits = firm.getMessageOfType(Messages.DepositsMessage.class).cash;
            }
        });
    }

    public static Action<Firm> receiveBailoutPackage() {
        return Action.create(Firm.class, firm -> {
            firm.getMessagesOfType(Messages.BailoutPackageMessage.class).forEach(msg -> {
                firm.price = msg.price;
                firm.wage = msg.wage;
                firm.deposits = 0;
            });
        });
    }

    public static Action<Firm> receiveBankruptcyMessage() {
        return Action.create(Firm.class, firm -> {
            firm.getMessagesOfType(Messages.BankruptcyMessage.class).forEach(msg -> {
                firm.deposits = 0;
                firm.productivity = false;
                firm.output = 0;
            });
        });
    }

    public static Action<Firm> doRevival() {
        return Action.create(Firm.class, firm -> {
            if (!firm.productivity & (firm.getPrng().uniform(0, 1).sample() < firm.getGlobals().phi)) {
                firm.productivity = true;
                firm.price = firm.averagePrice;
                firm.output = firm.getGlobals().mu * firm.unemployment * firm.getPrng().uniform(0, 1).sample();
                firm.deposits = firm.output * firm.wage;
                firm.debt = firm.deposits;
            }
            if (firm.productivity & firm.deposits > 0) {
                firm.cash = firm.deposits;
            }
        });
    }

    public static Action<Firm> sendDebt() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.DebtMessage.class, (msg, link) -> {
                msg.debt = firm.debt;
            });
            firm.debt = 0;
        });
    }

    public static Action<Firm> sendCash() {
        return Action.create(Firm.class, firm -> {
            firm.getLinks(Links.FirmToEconomyLink.class).send(Messages.CashMessage.class, (msg, link) -> {
                msg.cash = firm.cash;
            });
            firm.cash = 0;
        });
    }

    public static Action<Firm> receiveCash() {
        return Action.create(Firm.class, firm -> {
            if (firm.getMessagesOfType(Messages.CashMessage.class).stream().count() > 0) {
                firm.deposits = firm.getMessageOfType(Messages.CashMessage.class).cash;
            }
        });
    }
}
